import inspect
from copy import copy

class CalveMachine(object):

    def __init__(self):
        self.seeds = {}
        self.seed_inheritance = {}

    def calve(self, seed_name, **acquired_features):
        seed_hierarchy = self._build_hierarchy(seed_name)
        ovum = _Ovum(seed_hierarchy, acquired_features,
            acquired_features.pop('template', None))
        return ovum.generate_child()

    def _build_hierarchy(self, seed_name):
        this_seed = self.seeds[seed_name]
        seed_hierarchy = [this_seed]
        while this_seed in self.seed_inheritance:
            ancient_seed = self.seed_inheritance[this_seed]
            seed_hierarchy.insert(0, ancient_seed)
            this_seed = ancient_seed
        return seed_hierarchy

    def _receive(self, seed, seed_from=None):
        self.seeds[seed.__name__] = seed
        if seed_from:
            self.seed_inheritance[seed] = self.seeds[seed_from]


class _Ovum(object):

    def __init__(self, seed_hierarchy, acquired_features, cloned_by=None):
        self.__dict__['eaten'] = {}
        self.__dict__['cloned_by'] = cloned_by
        self.__dict__['seeds'] = seed_hierarchy
        self.__dict__['acquired_features'] = acquired_features
        self.__dict__['_after_build'] = lambda obj: 'do nothing by default'
        self.__dict__['init_args'] = []
        self.__dict__['init_kwargs'] = {}

    def after_build(self, block):
        self.__dict__['_after_build'] = block

    def init(self, *args, **kwargs):
        self.__dict__['init_args'] = args
        self.__dict__['init_kwargs'] = kwargs

    def __setattr__(self, attrname, value):
        self.eaten[attrname] = food = self._make_nutrient(attrname, value)
        setattr(food, 'attribute_name', attrname)

    def _make_nutrient(self, attribute_name, food):
        if isinstance(food, Nutrient):
            food.attribute_name = attribute_name
            return food
        else:
            return TransparentNutrient(food)

    def generate_child(self):
        self._fertilize()
        child = self._build_child()
        self._after_build(child)
        return child

    def _fertilize(self):
        for seed in self.seeds:
            seed(self)

    def _build_child(self):
        child = self._create_tabula_rasa()
        self._apply_inherited_features(child)
        self._apply_acquired_features(child)
        return child

    def _create_tabula_rasa(self):
        return self._dna() or self._blank_slate()

    def _apply_inherited_features(self, child):
        for attrname, value in self.eaten.items():
            setattr(child, attrname, value.run())

    def _apply_acquired_features(self, child):
        for field, value in self.acquired_features.items():
            setattr(child, field, value)

    def _dna(self):
        dna, args, kwargs = self.cloned_by, self.init_args, self.init_kwargs
        return inspect.isclass(dna) and self._build_dna() or copy(dna)

    def _blank_slate(self):
        return type('BlankSlate', (object,), {})()

    def _build_dna(self):
        try:
            return self.cloned_by(*self.init_args, **self.init_kwargs)
        except TypeError:
            return self.cloned_by()


class Nutrient(object):
    pass


class sequence(Nutrient):

    _indices = {}

    def __init__(self, block):
        self._block = block

    @classmethod
    def _next(cls, attribute_name):
        if attribute_name not in cls._indices:
            cls._indices[attribute_name] = 0
        cls._indices[attribute_name] += 1
        return cls._indices[attribute_name]

    def next(self):
        return self.run()

    def run(self):
        return self._block(sequence._next(self.attribute_name))


class TransparentNutrient(Nutrient):

    def __init__(self, food):
        self.food = food

    def run(self):
        return self.food


class association(Nutrient):

    def __init__(self, seed_name=None):
        self.seed_name = seed_name

    def run(self):
        seed_name = self.seed_name or self.attribute_name
        return pregnant.calve(seed_name)


_woman = pregnant = CalveMachine()

def inseminate(*seed, **options):
    if seed:
        _woman._receive(seed[0])
    else:
        def decorator(seed):
            _woman._receive(seed, seed_from=options['seed_from'])
        return decorator

